/*********************************************************************
*
*   MODULE NAME:
*       CCdtChatDlg.cpp
*
* Copyright 2014 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/
#include "stdafx.h"
#include "CCdtChatDlg.h"
#include "comdef.h"

#if( CDT_SUPPORT )

// the limit imposed on a single message, which will get separated into
// chunks if greater than packet payload length
#define MAX_INPUT_LENGTH    1024

// CCdtChatDlg dialog

IMPLEMENT_DYNAMIC(CCdtChatDlg, CDialog)

BEGIN_MESSAGE_MAP(CCdtChatDlg, CDialog)
    ON_WM_SIZE()
    ON_MESSAGE( WM_EVENT( EVENT_CDT_PACKET_RECEIVED ), OnCDTPacketReceived )
END_MESSAGE_MAP()

//----------------------------------------------------------------------
//! \brief Constructor
//----------------------------------------------------------------------
CCdtChatDlg::CCdtChatDlg(CWnd* pParent, FmiApplicationLayer & aCom)
    : CDialog( CCdtChatDlg::IDD, pParent ),
    mCom( aCom )
    {
    }

//----------------------------------------------------------------------
//! \brief Destructor
//----------------------------------------------------------------------
CCdtChatDlg::~CCdtChatDlg()
{
    CoUninitialize();
}

//----------------------------------------------------------------------
//! \brief Perform dialog data exchange and validation
//! \param  aDataExchange The DDX context
//----------------------------------------------------------------------
void CCdtChatDlg::DoDataExchange(CDataExchange* pDX)
   {
   CDialog::DoDataExchange(pDX);
   DDX_Control( pDX, IDC_CDT_CHAT_LOG, mBrowser );
   DDX_Control( pDX, IDC_CDT_CHAT_INPUT, mInputField );
   }

//----------------------------------------------------------------------
//! \brief This function is called when the window is created.
//! \return TRUE, since this function does not set focus to a control
//----------------------------------------------------------------------
BOOL CCdtChatDlg::OnInitDialog()
    {
    CoInitialize(NULL);

    CDialog::OnInitDialog();

    mInputField.SetLimitText( MAX_INPUT_LENGTH );

    CRect rect1, rect2;
    mBrowser.GetClientRect( &rect1 );
    GetClientRect( &rect2 );
    mPadding = ( rect2.Width() - rect1.Width() ) / 2;
    mInputField.GetClientRect( &rect1 );
    mInputHeight = rect1.Height();

    _variant_t    flags( 0L, VT_I4 );
    _variant_t    target_frame_name( "" );
    _variant_t    post_data( "" );
    _variant_t    headers( "" );
    mBrowser.Navigate( _T( "about:blank" ), &flags, &target_frame_name, &post_data, &headers );
    //navigating to chat-starter.html will not work because later using browser.WriteContent will
    //overwrite the chat-starter.html content and therefore lose all styling
    CFile starterFile;
    if ( starterFile.Open( _T( "cdt\\chat-starter.html" ), CFile::modeRead ) )
        {
        char buf[1024 + 1];
        UINT readLen;
        while( 0 < ( readLen = starterFile.Read( buf, 1024 ) ) )
            {
            buf[ readLen ] = 0;
            mBrowser.WriteContent( CString( buf ) );
            }
        starterFile.Close();
        }

    updateSize();

    return TRUE;
    }

//----------------------------------------------------------------------
//! \brief structure for chat messages and chat message reliability ACKs
//----------------------------------------------------------------------
__packed struct chat_msg_header
    {
    uint8  type;
    uint8  chunkId;
    uint8  reserved[2];  //byte alignment padding
    uint32 id;
    };

//the header structure is the same size and format as the full chat message
//structure up until the body buffer so that the header can be read to
//decide if the packet is a full message or not
__packed struct chat_msg
    {
    uint8  type;
    uint8  chunkId;
    uint8  bodyLength;
    uint8  reserved; //byte alignment padding
    uint32 id;
    //for simplicity of processing, ensure that this holds an even number
    uint8  body[ MAX_CDT_PAYLOAD - sizeof( chat_msg_header ) ];
    };

static COleDateTime epoch(1970,1,1,0,0,0);

// CCdtChatDlg message handlers
//----------------------------------------------------------------------
//! \brief Click handler for OK button
//! \details Build and send a SET baud rate packet based on the state
//!  of the fields
//----------------------------------------------------------------------
void CCdtChatDlg::OnOK()
{
    //should get here by using Enter key in input field or by clicking Send
    if( 0 == mInputField.GetWindowTextLengthW() )
        {
        return;
        }

    TCHAR text[ MAX_INPUT_LENGTH + 1 ];
    mInputField.GetWindowText( text, sizeof( text ) );
    int byteLength = _tcslen( text ) * sizeof( TCHAR );
    uint8 converted[ MAX_INPUT_LENGTH * sizeof( TCHAR ) ];
    memset( converted, 0, sizeof( converted ) );
    memcpy( converted, (void*)text, byteLength );
    chat_msg payload;
    payload.type = 0;
    COleDateTime currentTime = COleDateTime::GetCurrentTime();
    COleDateTimeSpan diff = currentTime - epoch;
    payload.id = diff.GetTotalSeconds();
    payload.chunkId = 0;
    payload.reserved = 0;

    CString formatted;
    formatted.Format(
        _T( "<div class='sent' id='%u'><span class='ts'>%s</span>" ),
        payload.id,
        currentTime.Format( _T( "%Y-%m-%d %H.%M.%S" ) )
        );
    TRACE1( "sent chat message %u\n", payload.id );

    TCHAR chunk[ MAX_CDT_PAYLOAD ];

    for( int i=0; i<byteLength; i += sizeof( payload.body ) )
        {
        memset( &payload.body, 0, sizeof( payload.body ) );
        payload.bodyLength = min( byteLength - i, sizeof( payload.body ) );
        memcpy( &payload.body, &converted[i], payload.bodyLength );
        mCom.txCdt( CDT_DATA_TRANSFER, (uint8*)&payload, sizeof( chat_msg_header ) + payload.bodyLength );

        memset( chunk, 0, sizeof( chunk ) );
        memcpy( chunk, &converted[i], payload.bodyLength );
        formatted.AppendFormat(
            _T( "<span class='chunk' chunk-id='%d'>%s</span>" ),
            payload.chunkId,
            chunk
            );

        payload.chunkId++;
        }
    mBrowser.WriteContent( formatted );
    mBrowser.WriteContent( _T( "</div><div class='clear'></div>\n" ) );

    mInputField.SetWindowTextW( _T( "" ) );
    mInputField.SetFocus();
}

//----------------------------------------------------------------------
//! \brief This function handles inbound CDT packets
//----------------------------------------------------------------------
LRESULT CCdtChatDlg::OnCDTPacketReceived( WPARAM wParam, LPARAM lParam)
    {
    //TODO if it was a chat packet, then display
    sint16 cdtPayloadSize = (sint16) wParam;
    uint8 const * cdtPayload = (uint8 const *) lParam;

    //must have at least the size of a chat message header to be handled here
    if( cdtPayloadSize < sizeof( chat_msg_header ) )
        {
        return 0;
        }

    chat_msg packet;
    memset( &packet, 0, sizeof( chat_msg ) );
    memcpy( &packet, cdtPayload, cdtPayloadSize );

    switch( packet.type )
        {
        case 0:
            {
            //ACK the message
            chat_msg_header ack;
            memset( &ack, 0, sizeof( chat_msg_header ) );
            ack.type = 1;
            ack.id = packet.id;
            ack.chunkId = packet.chunkId;
            mCom.txCdt( CDT_DATA_TRANSFER, (uint8*)&ack, sizeof( chat_msg_header ) );

            //display in chat log
            TCHAR text[ MAX_CDT_PAYLOAD + 1 ];
            memset( text, 0, sizeof( text ) );
            memcpy( (void*)text, (void*)&packet.body, cdtPayloadSize - sizeof( chat_msg_header ) );
            CString formatted;
            formatted.Format(
                _T( "<div class='received' received-msg-id='%u'><span class='ts'>%s</span><span class='body'>%s</span></div><div class='clear'></div>\n" ),
                packet.id,
                COleDateTime::GetCurrentTime().Format( _T( "%Y-%m-%d %H.%M.%S" ) ),
                text
                );
            mBrowser.WriteContent( formatted );
            break;
            }
        case 1:
            {
            TRACE1( "received ACK for sent chat message %u\n", packet.id );
            CStringArray args;
            CString sentMessageId;
            sentMessageId.Format( _T( "%u" ), packet.id );
            args.Add( sentMessageId );
            CString chunkId;
            chunkId.Format( _T( "%d" ), packet.chunkId );
            args.Add( chunkId );
            mBrowser.InvokeScript( _T( "setDelivered" ), &args );
            break;
            }
        }

    return 0;
    }

//----------------------------------------------------------------------
//! \brief Called after the dialog is resized; repositions the contents
//!     of the display.
//! \details Moves/sizes the contents as follows:
//!     - Moves the search box to the bottom-left corner, keeping its
//!       size
//!     - Moves the packet list to the top-left, expands vertically
//!     - Moves the packet detail to the right of the list, expands
//!       horizontally to fill the dialog and vertically to match the
//!       height of the packet list
//!     - Centers the buttons under the packet detail.
//! \param aType The type of resizing requested (maximized, etc.)
//! \param aClientWidth The new width of the client area
//! \param aClientHeight The new height of the client area
//----------------------------------------------------------------------
void CCdtChatDlg::OnSize
    (
    UINT aType,
    int  aClientWidth,
    int  aClientHeight
    )
{
    CDialog::OnSize( aType, aClientWidth, aClientHeight );

    if ( NULL == mBrowser.GetSafeHwnd() )
        {
        return;
        }

    updateSize();
    }

//----------------------------------------------------------------------
//! \brief repositions the contents of the display.
//----------------------------------------------------------------------
void CCdtChatDlg::updateSize()
    {
    CRect rect;
    GetClientRect( &rect );

    mBrowser.MoveWindow(
        rect.left + mPadding,
        rect.top + mPadding,
        rect.Width() - ( 2 * mPadding ),
        rect.Height() - ( ( 3 * mPadding ) + mInputHeight )
        );
    mInputField.MoveWindow(
        rect.left + mPadding,
        rect.Height() - ( mPadding + mInputHeight ),
        rect.Width() - ( ( 3 * mPadding ) + mInputHeight),
        mInputHeight
        );
    GetDlgItem( IDOK )->MoveWindow(
        rect.right - ( mPadding + mInputHeight ),
        rect.Height() - ( mPadding + mInputHeight ),
        mInputHeight,
        mInputHeight
        );
}

#endif