/*********************************************************************
*
*   MODULE NAME:
*       CDashcamNtfctnDlg.cpp
*
* Copyright 2015 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/
#include "stdafx.h"
#include "CDashcamNtfctnDlg.h"
#include "Event.h"

#if( FMI_SUPPORT_A622 )

// CDashcamNtfctn dialog

IMPLEMENT_DYNAMIC(CDashcamNtfctnDlg, CDialog)

BEGIN_MESSAGE_MAP(CDashcamNtfctnDlg, CDialog)
    ON_MESSAGE( WM_EVENT( EVENT_FMI_DASHCAM_SET_NTFCTN_RECEIPT ), OnDashcamSetNtfctnReceipt )
    ON_MESSAGE( WM_EVENT( EVENT_FMI_DASHCAM_GET_NTFCTN_SETTINGS_RESPONSE ), OnGetNtfctnSettingsResponse )
    ON_BN_CLICKED( IDC_ENABLE_FORWARD_COLLISION_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_DISABLE_FORWARD_COLLISION_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_ENABLE_LANE_DEPARTURE_WARNING_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_DISABLE_LANE_DEPARTURE_WARNING_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_ENABLE_INCIDENT_DETECTION_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_DISABLE_INCIDENT_DETECTION_NTFCTN, OnReevaluate )
    ON_BN_CLICKED( IDC_DASHCAM_REFRESH_SETTINGS, OnRefreshNtfctnSettings )
    ON_BN_CLICKED( IDOK, OnBnClickedSet )
    ON_BN_CLICKED( IDCANCEL, OnClose )
END_MESSAGE_MAP()

//----------------------------------------------------------------------
//! \brief Constructor
//! \param aParent The parent of this dialog
//! \param aCom Reference to the FMI communication controller
//----------------------------------------------------------------------
CDashcamNtfctnDlg::CDashcamNtfctnDlg
    (
    CWnd                * aParent,
    FmiApplicationLayer & aCom
    )
    : CDialog( CDashcamNtfctnDlg::IDD, aParent )
    , mCom( aCom )
    , mForwardCollisionNtfctn( -1 )
    , mLaneDepartureNtfctn( -1 )
    , mIncidentDetectionNtfctn( -1 )
{
}

//----------------------------------------------------------------------
//! \brief Destructor
//----------------------------------------------------------------------
CDashcamNtfctnDlg::~CDashcamNtfctnDlg()
{
}

//----------------------------------------------------------------------
//! \brief Initialize the dialog
//! \details This function is called when the window is created. It
//!     sets up the parent, so it can get info from and send a message
//!     to FmiApplicationLayer.
//! \return TRUE, since this function does not set focus to a control
//----------------------------------------------------------------------
BOOL CDashcamNtfctnDlg::OnInitDialog()
{
    CDialog::OnInitDialog();

    UpdateData( FALSE );
    return TRUE;
}   /* OnInitDialog() */

//----------------------------------------------------------------------
//! \brief Perform dialog data exchange and validation
//! \param  aDataExchange The DDX context
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::DoDataExchange
    (
    CDataExchange * aDataExchange
    )
{
    CDialog::DoDataExchange( aDataExchange );

    DDX_Radio( aDataExchange, IDC_DISABLE_FORWARD_COLLISION_NTFCTN, mForwardCollisionNtfctn );
    DDX_Radio( aDataExchange, IDC_DISABLE_LANE_DEPARTURE_WARNING_NTFCTN, mLaneDepartureNtfctn );
    DDX_Radio( aDataExchange, IDC_DISABLE_INCIDENT_DETECTION_NTFCTN, mIncidentDetectionNtfctn );
}

//----------------------------------------------------------------------
//! \brief Button handler for the Retrieve Settings button
//! \details Calls FmiApplicationLayer to initiate the Get Ntfctn settings
//!     request protocol
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::OnRefreshNtfctnSettings()
{
    UpdateData( TRUE );
    fmi_dashcam_get_settings_request_data_type getSettings;
    getSettings.change_id = GetTickCount();
    mCom.sendFmiPacket( FMI_DASHCAM_GET_NTFCTN_SETTINGS_REQUEST, (uint8*)&getSettings, sizeof( getSettings) );
    UpdateData( FALSE );
}

//----------------------------------------------------------------------
//! \brief Button handler for the Set button
//! \details Retrieves configuration values from user interface and
//!         sends dashcam notification request protocol
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::OnBnClickedSet()
{
    fmi_dashcam_set_ntfctn_settings_data_type setNtfctn;
    memset( &setNtfctn, 0, sizeof( setNtfctn ) );
    setNtfctn.change_id = GetTickCount();

    setNtfctn.enable_fcw_ntfctn = (boolean) mForwardCollisionNtfctn;
    setNtfctn.enable_incident_ntfctn = (boolean) mIncidentDetectionNtfctn;
    setNtfctn.enable_lane_departure_ntfctn = (boolean) mLaneDepartureNtfctn;

    mCom.sendFmiPacket( FMI_DASHCAM_SET_NTFCTN_REQUEST, (uint8*)&setNtfctn, sizeof( setNtfctn ) );
}    /* OnBnClickedSet () */

//----------------------------------------------------------------------
//! \brief Handler for the Close action
//! \details Closes the window.
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::OnClose()
{
    UpdateData( TRUE );
    mForwardCollisionNtfctn = -1;
    mLaneDepartureNtfctn = -1;
    mIncidentDetectionNtfctn = -1;
    UpdateData( FALSE );

    CDialog::OnCancel();
}    /* OnClose() */

//----------------------------------------------------------------------
//! \brief Called by MFC after the window has been destroyed; performs
//!     final termination activities.
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::PostNcDestroy()
{
    CDialog::PostNcDestroy();
}   /* PostNcDestroy() */


//----------------------------------------------------------------------
//! \brief This function handles the dashcam set notification receipt event from device
//----------------------------------------------------------------------
LRESULT CDashcamNtfctnDlg::OnDashcamSetNtfctnReceipt ( WPARAM return_code, LPARAM )
{
    CString message;

    switch( return_code )
    {
    case DASHCAM_SUCCESS:
        {
        message.Format( _T( "Camera notification settings successfully updated." ) );
        MessageBox( message, _T( "Camera Notifications Success" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    case DASHCAM_ERR_FCW_LDW_NTFCTN_NOT_SUPPORT:
        {
        message.Format( _T( "Camera notifications failed.\n Forward Collision Warnings and Lane Departure Warnings not supported on this device." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    case DASHCAM_ERR_FCW_NOT_ENABLED:
        {
        message.Format( _T( "Camera notifications failed. Forward Collision Warning is not enabled on this device." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    case DASHCAM_ERR_LDW_NOT_ENABLED:
        {
        message.Format( _T( "Camera notifications failed. Lane Departure Warning is not enabled on this device." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    case DASHCAM_ERR_INCIDENT_NOT_ENABLED:
        {
        message.Format( _T( "Camera notifications failed. Automatic Incident Detection is not enabled on this device." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    default:
        {
        message.Format( _T( "Camera notifications failed. Unrecognized error." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    }
    return 0;
}

//----------------------------------------------------------------------
//! \brief This function handles the dashcam set notification receipt event from device
//----------------------------------------------------------------------
LRESULT CDashcamNtfctnDlg::OnGetNtfctnSettingsResponse ( WPARAM return_code, LPARAM receipt_data )
{
    CString message;
    fmi_dashcam_get_ntfctn_settings_data_type messageEvent = *( ( fmi_dashcam_get_ntfctn_settings_data_type* ) receipt_data );

    switch( return_code )
    {
    case DASHCAM_SUCCESS:
        {
        UpdateData( TRUE );
        mForwardCollisionNtfctn = messageEvent.enable_fcw_ntfctn;
        mLaneDepartureNtfctn = messageEvent.enable_lane_departure_ntfctn;
        mIncidentDetectionNtfctn = messageEvent.enable_incident_ntfctn;
        UpdateData( FALSE );

        message.Format( _T( "Camera notification settings successfully retrieved." ) );
        MessageBox( message, _T( "Camera Notifications Success" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    case DASHCAM_ERR_INTERNAL:
        {
        message.Format( _T( "Camera notification settings retrieval failed due to internal error." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    default:
        {
        message.Format( _T( "Camera notification settings retrieval failed. Unrecognized Reason." ) );
        MessageBox( message, _T( "Camera Notifications Failure" ), MB_OK | MB_ICONINFORMATION );
        break;
        }
    }
    return 0;
}

//----------------------------------------------------------------------
//! \brief Change handler for editable controls
//! \details Make the SET button clickable if the data is valid for sending
//----------------------------------------------------------------------
void CDashcamNtfctnDlg::OnReevaluate()
{
    UpdateData();
    //checks to make sure all notifications are set
    //if so, set button is enabled, if not, set button disabled.
    if( ( -1 != mForwardCollisionNtfctn ) &&
        ( -1 != mLaneDepartureNtfctn ) &&
        ( -1 != mIncidentDetectionNtfctn ) )
    {
        GetDlgItem( IDOK )->EnableWindow();
    }
    else
    {
        GetDlgItem( IDOK )->EnableWindow( FALSE );
    }
}

#endif