/*********************************************************************
*
*   MODULE NAME:
*       CFileConverterDlg.cpp
*
*   Copyright 2008-2016 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/
#include "stdafx.h"
#include "Shlwapi.h"

#include <fstream>
#include <sstream>
#include <string>
#include <vector>

#include "CFileConverterDlg.h"
#include "util.h"

using namespace std;

#if( FMI_SUPPORT_A622 )

IMPLEMENT_DYNAMIC( CFileConverterDlg, CDialog )

BEGIN_MESSAGE_MAP( CFileConverterDlg, CDialog )
    ON_EN_CHANGE( IDC_FILE_CONVERT_EDIT_SRC_FILENAME, OnEnChangeFile )
    ON_EN_CHANGE( IDC_FILE_CONVERT_EDIT_DST_FILENAME, OnEnChangeFile )
    ON_EN_CHANGE( IDC_FILE_CONVERT_EDIT_DST_FOLDER, OnEnChangeFile )
    ON_EN_CHANGE( IDC_FILE_CONVERT_EDIT_STOPID, OnEnChangeFile )
    ON_EN_CHANGE( IDC_FILE_CONVERT_EDIT_ROUTE_NAME, OnEnChangeFile )
    ON_CBN_SELENDOK( IDC_FILE_CONVERT_CBO_TYPE, OnComboChanged )
    ON_BN_CLICKED( IDC_FILE_CONVERT_BTN_FIND, OnBnClickedFind )
    ON_BN_CLICKED( IDC_FILE_CONVERT_BTN_CHOOSE, OnBnClickedChoose )
    ON_BN_CLICKED( IDOK, OnBnClickedOk )
END_MESSAGE_MAP()

//! Valid values for file conversion types
enum conversion_type_type
{
    CONV_PSR_TO_CSV = 0,
    CONV_CSV_TO_PSR,
    CONV_GPX_TO_PSR
};

/*--------------------------------------------------------------------
LITERAL CONSTANTS
--------------------------------------------------------------------*/

//! Conversion types that the client accepts
static CFileConverterDlg::conversion_type conversion_types[] =
{
    { CONV_PSR_TO_CSV, _T( "PSR to CSV" ), NULL, _T(".csv"), _T( "PSR Files (*.*)|*.*|" ) },
    { CONV_CSV_TO_PSR, _T( "CSV to PSR" ), NULL, _T(".psr"), _T( "CSV Files (*.csv)|*.csv|" ) },
    { CONV_GPX_TO_PSR, _T( "GPX to PSR" ), NULL, _T(".psr"), _T( "GPX Files (*.gpx)|*.gpx|" ) }
};

//----------------------------------------------------------------------
//! \brief Constructor
//! \param aParent The parent of this dialog
//! \param aCom Reference to the FMI communication controller
//----------------------------------------------------------------------
CFileConverterDlg::CFileConverterDlg
    (
    CWnd                * aParent,
    FmiApplicationLayer & aCom
    )
    : CDialog( CFileConverterDlg::IDD, aParent )
    , mCom( aCom )
    , mConversionType( 0 )
    , mSrcFilePath( _T("") )
    , mDstFileName( _T("") )
    , mDstFolder( _T("") )
    , mConversionIndex( 0 )
    , mStopId( _T("") )
    , mRouteName( _T("") )
    , mSrcCoords( 0 )
    , mDstCoords( 0 )
{
}

//----------------------------------------------------------------------
//! \brief Destructor
//----------------------------------------------------------------------
CFileConverterDlg::~CFileConverterDlg()
{
}

//----------------------------------------------------------------------
//! \brief Perform dialog data exchange and validation
//! \param aDataExchange The DDX context
//----------------------------------------------------------------------
void CFileConverterDlg::DoDataExchange
    (
    CDataExchange * aDataExchange
    )
{
    CDialog::DoDataExchange( aDataExchange );

    DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_SRC_FILENAME, mSrcFilePath );
    DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_DST_FILENAME, mDstFileName );
    DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_DST_FOLDER, mDstFolder );

    if( mConversionType == CONV_CSV_TO_PSR )
    {
        DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_STOPID, mStopId );
        DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_ROUTE_NAME, mRouteName );
    }
    else if( mConversionType == CONV_GPX_TO_PSR )
    {
        DDX_Text( aDataExchange, IDC_FILE_CONVERT_EDIT_STOPID, mStopId );
    }
}

//----------------------------------------------------------------------
//! \brief Initialize the dialog
//! \details This function is called when the window is created. It
//!     sets up the parent, so it can get info from and send a message
//!     to FmiApplicationLayer.
//! \return TRUE, since this function does not set focus to a control
//----------------------------------------------------------------------
BOOL CFileConverterDlg::OnInitDialog()
{
    CDialog::OnInitDialog();

    // Populate file type combobox
    CComboBox * cbo = (CComboBox *)GetDlgItem( IDC_FILE_CONVERT_CBO_TYPE );
    cbo->ResetContent();
    for( int i = 0; i < cnt_of_array( conversion_types ); i++ )
    {
        cbo->AddString( conversion_types[i].text );
    }
    cbo->SetCurSel( 0 );
    mConversionType = conversion_types[0].index;

    // Initially set the source file radio button to "Semi-Circles"
    CheckRadioButton( IDC_FILE_CONVERT_RADB_1, IDC_FILE_CONVERT_RADB_2, IDC_FILE_CONVERT_RADB_2 );
    // Set the destination file radio button to "Degrees"
    CheckRadioButton( IDC_FILE_CONVERT_RADB_3, IDC_FILE_CONVERT_RADB_4, IDC_FILE_CONVERT_RADB_3 );
    GetDlgItem( IDC_FILE_CONVERT_RADB_1 )->EnableWindow( FALSE );
    GetDlgItem( IDC_FILE_CONVERT_RADB_2 )->EnableWindow( FALSE );

    UpdateData( FALSE );
    return TRUE;
} /* OnInitDialog() */

//----------------------------------------------------------------------
//! \brief Click handler for OK button.
//! \details Initiates the file conversion process.  The type of file
//!  conversion is selected by the user. A post-processing popup alerts
//!  the user if the conversion was successful or unsuccessful.
//----------------------------------------------------------------------
void CFileConverterDlg::OnBnClickedOk()
{
    UpdateData( TRUE );

    // Generate the full destination file path
    CString dstFilePath;
    BOOL result = FALSE;

    dstFilePath = mDstFolder;
    if( !( dstFilePath.Right( 1 ) == _T("\\") ) )
    {
        dstFilePath.Append( _T("\\") );
    }
    dstFilePath.Append( mDstFileName );

    // Check if the destination file already exists.
    // If so, prompt the user to make sure they want to overwrite.
    CFileStatus status;
    if( CFile::GetStatus( dstFilePath, status ) != 0 )
    {
        if( MessageBox( _T("Destination File already exists. Overwrite File?"), _T("Caution"), MB_YESNO ) == IDNO )
        {
            return;
        }
    }
    if( IsDlgButtonChecked( IDC_FILE_CONVERT_RADB_1 ) == BST_CHECKED )
    {
        mSrcCoords = 1; // Degrees
    }
    else
    {
        mSrcCoords = 2; // Semi-circles
    }
    if( IsDlgButtonChecked( IDC_FILE_CONVERT_RADB_3 ) == BST_CHECKED )
    {
        mDstCoords = 1; // Degrees
    }
    else
    {
        mDstCoords = 2; // Semi-circles
    }

    // Setup Child Process STDOUT Redirection
    SECURITY_ATTRIBUTES saAttr;
    saAttr.nLength = sizeof( SECURITY_ATTRIBUTES );
    saAttr.bInheritHandle = TRUE;
    saAttr.lpSecurityDescriptor = NULL;

    // Create a pipe for the child process's STDOUT.
    if ( ! CreatePipe( &hChildStd_OUT_Rd, &hChildStd_OUT_Wr, &saAttr, 0 ) ) return;

    // Ensure the read handle to the pipe for STDOUT is not inherited.
    if ( ! SetHandleInformation( hChildStd_OUT_Rd, HANDLE_FLAG_INHERIT, 0 ) ) return;

    result = CreateChildProcess( dstFilePath );

    if( result == TRUE )
    {
        CString output;
        output.Format( _T("Successfully Converted the File.\n\nFile sent to:\n%s"),
            (LPCTSTR)dstFilePath );
        MessageBox( output, _T("Conversion Complete") );
    }
}

//----------------------------------------------------------------------
//! \brief Helper function to format the command line arguments
//! \details
//----------------------------------------------------------------------
CString CFileConverterDlg::FormatArguments( CString dstFilePath )
{
    CString arguments;

    // Append quotes around the file paths or route name in case of
    // spaces in the strings.
    mSrcFilePath.Insert( 0,'\"' );
    mSrcFilePath.AppendChar( '\"' );
    dstFilePath.Insert( 0, '\"' );
    dstFilePath.AppendChar( '\"' );
    mRouteName.Insert( 0, '\"' );
    mRouteName.AppendChar( '\"' );

    // Conversion type is 1-based in FleetFileUtility
    switch( mConversionType )
    {
        case CONV_PSR_TO_CSV:
            arguments.Format( _T("%i "), CONV_PSR_TO_CSV+1 );
            arguments.Append( mSrcFilePath );
            arguments.Append( _T(" "));
            arguments.Append( dstFilePath );
            arguments.Append( _T(" -D ") );
            arguments.AppendFormat( _T("%i"), mDstCoords );
            break;
        case CONV_CSV_TO_PSR:
            arguments.Format( _T("%i "), CONV_CSV_TO_PSR+1 );
            arguments.Append( mSrcFilePath );
            arguments.Append( _T(" "));
            arguments.Append( dstFilePath );
            arguments.Append( _T(" -R ") );
            arguments.Append( mRouteName );
            arguments.Append( _T(" -I ") );
            arguments.Append( mStopId );
            arguments.Append( _T(" -S ") );
            arguments.AppendFormat( _T("%i"), mSrcCoords );
            break;
        case CONV_GPX_TO_PSR:
            arguments.Format( _T("%i "), CONV_GPX_TO_PSR+1 );
            arguments.Append( mSrcFilePath );
            arguments.Append( _T(" "));
            arguments.Append( dstFilePath );
            arguments.Append( _T(" -I ") );
            arguments.Append( mStopId );
            break;
        default:
            arguments.Append( _T("-H") );
            break;
    }
    return arguments;
}

//----------------------------------------------------------------------
//! \brief Click handler for Find (file to convert) button.
//! \details Displays a File..Open dialog allowing the user to select
//!     the file to convert.
//----------------------------------------------------------------------
void CFileConverterDlg::OnBnClickedFind()
{
    UpdateData();

    TCHAR workingDirectory[200];
    // opening a file in another directory changes the current
    // directory, which will cause problems because the log and data
    // files are opened relative to the current directory.  So, get the
    // directory now so that it can be restored when the user is done
    // picking a file.
    DWORD result = GetCurrentDirectory( 200, workingDirectory );
    if( result == 0 || result > 200 )
    {
        MessageBox( _T("Unable to get current directory"), _T("Severe Error") );
        OnCancel();
        return;
    }
    CFileDialog dlg
        (
        TRUE,
        conversion_types[ mConversionIndex ].extension,
        NULL,
        OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
        conversion_types[ mConversionIndex ].filter
        );
    if( dlg.DoModal() == IDOK )
    {
        mSrcFilePath = dlg.GetPathName();
        mDstFileName = dlg.GetFileTitle();
        mDstFileName.Append( conversion_types[ mConversionIndex ].newExtension );
        mDstFolder = dlg.GetFolderPath();
        UpdateData( FALSE );
        OnEnChangeFile();
    }
    if( !SetCurrentDirectory( workingDirectory ) )
    {
        MessageBox( _T("Unable to set current directory"), _T("Severe Error") );
        OnCancel();
        return;
    }
}

//----------------------------------------------------------------------
//! \brief Click handler for Choose (destination folder) button.
//! \details Lets the user select a folder location to save the
//!     converted file.
//----------------------------------------------------------------------
void CFileConverterDlg::OnBnClickedChoose()
{
    UpdateData();

    // Archaic implementation. If this project is ever upgraded to VS2010+,
    //  update this code to use CFolderPickerDialog.
    LPMALLOC pMalloc;
    BROWSEINFO bi;
    ZeroMemory( &bi, sizeof( bi ) );
    TCHAR szDisplayName[MAX_PATH];
    szDisplayName[0]  = NULL;
    LPITEMIDLIST pidl;

    bi.hwndOwner = ::GetDesktopWindow();
    bi.pidlRoot = NULL;
    bi.pszDisplayName = szDisplayName;
    bi.lpszTitle = _T( "Please select a folder:" );
    bi.ulFlags = BIF_USENEWUI|BIF_RETURNONLYFSDIRS;
    bi.lParam = NULL;
    bi.iImage = 0;
    OleInitialize(NULL);

    if ( ::SHGetMalloc( &pMalloc ) == NOERROR )
    {
        if( NULL != ( pidl = SHBrowseForFolder( &bi ) ) )
        {
             TCHAR szPathName[MAX_PATH];
             if( SHGetPathFromIDList( pidl, szPathName ) )
             {
                mDstFolder = szPathName;
                UpdateData( FALSE );
                OnEnChangeFile();
             }
        }
        pMalloc->Free(pidl);
    }
    pMalloc->Release();
}

//----------------------------------------------------------------------
//! \brief Change Handler for the combo box.
//! \details This is needed because all messages for combo box changing
//!     are sent before the text has acutally changed. This workaround
//!     addresses that issue.
//----------------------------------------------------------------------
void CFileConverterDlg::OnComboChanged()
{
    CComboBox * cbo = (CComboBox *)GetDlgItem( IDC_FILE_CONVERT_CBO_TYPE );
    mConversionIndex = cbo->GetCurSel();
    mConversionType = conversion_types[mConversionIndex].index;

    // Enable/Disable the stopId/route name edit boxes based on user selection
    if( mConversionType == CONV_CSV_TO_PSR )
    {
        GetDlgItem( IDC_FILE_CONVERT_EDIT_STOPID )->EnableWindow( TRUE );
        GetDlgItem( IDC_FILE_CONVERT_EDIT_ROUTE_NAME )->EnableWindow( TRUE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_1 )->EnableWindow( TRUE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_2 )->EnableWindow( TRUE );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_3, IDC_FILE_CONVERT_RADB_4, IDC_FILE_CONVERT_RADB_4 );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_1, IDC_FILE_CONVERT_RADB_2, IDC_FILE_CONVERT_RADB_1 );
        GetDlgItem( IDC_FILE_CONVERT_RADB_3 )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_4 )->EnableWindow( FALSE );
    }
    else if( mConversionType == CONV_GPX_TO_PSR )
    {
        GetDlgItem( IDC_FILE_CONVERT_EDIT_STOPID )->EnableWindow( TRUE );
        GetDlgItem( IDC_FILE_CONVERT_EDIT_ROUTE_NAME )->EnableWindow( FALSE );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_1, IDC_FILE_CONVERT_RADB_2, IDC_FILE_CONVERT_RADB_1 );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_3, IDC_FILE_CONVERT_RADB_4, IDC_FILE_CONVERT_RADB_4 );
        GetDlgItem( IDC_FILE_CONVERT_RADB_1 )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_2 )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_3 )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_4 )->EnableWindow( FALSE );
    }
    else
    {
        GetDlgItem( IDC_FILE_CONVERT_EDIT_STOPID )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_EDIT_ROUTE_NAME )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_3 )->EnableWindow( TRUE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_4 )->EnableWindow( TRUE );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_1, IDC_FILE_CONVERT_RADB_2, IDC_FILE_CONVERT_RADB_2 );
        CheckRadioButton( IDC_FILE_CONVERT_RADB_3, IDC_FILE_CONVERT_RADB_4, IDC_FILE_CONVERT_RADB_3 );
        GetDlgItem( IDC_FILE_CONVERT_RADB_1 )->EnableWindow( FALSE );
        GetDlgItem( IDC_FILE_CONVERT_RADB_2 )->EnableWindow( FALSE );
    }

    // Item changed, so clear the edit boxes
    CEdit * ced = (CEdit *)GetDlgItem( IDC_FILE_CONVERT_EDIT_SRC_FILENAME );
    ced->SetWindowTextW( _T("") );
    ced = (CEdit *)GetDlgItem( IDC_FILE_CONVERT_EDIT_DST_FILENAME );
    ced->SetWindowTextW( _T("") );
    ced = (CEdit *)GetDlgItem( IDC_FILE_CONVERT_EDIT_DST_FOLDER );
    ced->SetWindowTextW( _T("") );
    ced = (CEdit *)GetDlgItem( IDC_FILE_CONVERT_EDIT_STOPID );
    ced->SetWindowTextW( _T("") );
    ced = (CEdit *)GetDlgItem( IDC_FILE_CONVERT_EDIT_ROUTE_NAME );
    ced->SetWindowTextW( _T("") );

    UpdateData(FALSE);
}

//----------------------------------------------------------------------
//! \brief Edit Change handler for all text boxes.
//! \details Enables the OK button if all three edit controls are
//!     populated.
//----------------------------------------------------------------------
void CFileConverterDlg::OnEnChangeFile()
{
    BOOL isValid = TRUE;

    UpdateData( TRUE );

    if( mSrcFilePath == "" || mDstFileName == "" || mDstFolder == "" )
    {
        isValid = FALSE;
    }
    if( mConversionType == CONV_CSV_TO_PSR && ( mStopId == "" || mRouteName == "" ) )
    {
        isValid = FALSE;
    }
    else if( mConversionType == CONV_GPX_TO_PSR && mStopId == "" )
    {
        isValid = FALSE;
    }

    if( isValid )
    {
        GetDlgItem( IDOK )->EnableWindow( TRUE );
    }
    else
    {
        GetDlgItem( IDOK )->EnableWindow( FALSE );
    }
}

//----------------------------------------------------------------------
//! \brief Create a child process to execute the file conversion
//! \details Executes the FleetFileUtility executable. Any output from
//!     the tool is collected and displayed via message box. Much of
//!     the following is based on example code provided by MSDN.
//----------------------------------------------------------------------
BOOL CFileConverterDlg::CreateChildProcess( CString dstFilePath )
{
    PROCESS_INFORMATION piProcInfo;
    STARTUPINFO siStartInfo;
    BOOL bSuccess = FALSE;

//  Set up members of the PROCESS_INFORMATION structure.
    ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );

//  Set up members of the STARTUPINFO structure.
//  This structure specifies the STDOUT handles for redirection.
    ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
    siStartInfo.cb = sizeof(STARTUPINFO);
    siStartInfo.hStdError = hChildStd_OUT_Wr;
    //siStartInfo.hStdOutput = hChildStd_OUT_Wr;  // uncomment for additional debug output
    siStartInfo.dwFlags |= STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    siStartInfo.wShowWindow = SW_HIDE;

//  Create the command line arguments
    const int MAX_CMDLINE = 1000;
    WCHAR modulePath[MAX_PATH];
    ZeroMemory( &modulePath, MAX_PATH );
    WCHAR fullPathBuffer[MAX_CMDLINE];
    WCHAR exe[] = _T("\\FleetFileUtility.exe");  // The exe must be in the same location as FMC!!!
    GetModuleFileName( NULL, modulePath, MAX_PATH );
    PathRemoveFileSpec( modulePath );
    wcscat_s( modulePath, exe );
    wstring actualPath( modulePath );
    actualPath.insert(0, L"\""); // Enclose the exe path in quotes in case of spaces in the name
    actualPath.append(L"\" ");   // A trailing space is left to divide the path from the arguments
    CString parameters = FormatArguments( dstFilePath );
    actualPath.append( (wstring)parameters );
    wcscpy_s( fullPathBuffer, actualPath.size()+1, actualPath.c_str() );

//  Create the child process
    bSuccess = CreateProcess
                   ( modulePath         // Application name
                   , fullPathBuffer     // Full Command line
                   , NULL               // Process handle not inheritable
                   , NULL               // Thread handle not inheritable
                   , TRUE               // Handle inheritance true
                   , CREATE_NO_WINDOW   // Do not create a window
                   , NULL               // Use parent's environment
                   , NULL               // Use parent's directory
                   , &siStartInfo       // pointer to STARTUPINFO
                   , &piProcInfo        // pointer to PROCESS_INFORMATION
                   );

    if( bSuccess )
    {
        GetDlgItem( IDOK )->EnableWindow( FALSE ); // Disable the buttons in case of long conversion
        GetDlgItem( IDCANCEL )->EnableWindow( FALSE );
        // Wait for the process to finish running
        WaitForSingleObject( piProcInfo.hProcess, INFINITE );
        GetDlgItem( IDOK )->EnableWindow( TRUE ); // Re-enable
        GetDlgItem( IDCANCEL )->EnableWindow( TRUE );

        ReadFromPipe();
        DWORD prcRes;
        GetExitCodeProcess( piProcInfo.hProcess, &prcRes );
        bSuccess = (BOOL)prcRes;
    }
    else
    {
        MessageBox( _T("Unable to locate the Fleet File Utility. Make sure the")
                    _T(" FleetFileUtility.exe is in the same location as the Fleet")
                    _T("ManagementController.exe and try again.")
                    , _T("Severe Error") );
    }
    // Close the process handles
    CloseHandle( piProcInfo.hProcess );
    CloseHandle( piProcInfo.hThread );

    return bSuccess;
}

//----------------------------------------------------------------------
//! \brief Read output from the child process's pipe for STDOUT
//! \details Displays any of the tool's output in message box
//----------------------------------------------------------------------
void CFileConverterDlg::ReadFromPipe( void )
{
    const int BUFSIZE = 4096;
    DWORD dwRead;
    CHAR chBuf[BUFSIZE];
    ZeroMemory( chBuf, BUFSIZE );
    BOOL bSuccess = FALSE;

    CloseHandle( hChildStd_OUT_Wr ); // Reading from the stdout pipe is a blocking
                                     // call so we have to close the handle before
                                     // reading any output from the child process.
    bSuccess = ReadFile( hChildStd_OUT_Rd, chBuf, BUFSIZE, &dwRead, NULL );
    if( dwRead > 0 )
    {
        CString output( chBuf );
        MessageBox( output, _T( "Fleet File Utility" ) );
    }
}
#endif