/*********************************************************************
*
*   MODULE NAME:
*       FileConverter.cpp
*
* Copyright 2008-2016 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/

#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <algorithm>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/optional/optional.hpp>
#include <boost/foreach.hpp>
#include <boost/program_options/detail/utf8_codecvt_facet.hpp>

#include "FileConverter.h"

namespace po = boost::program_options;
namespace pt = boost::property_tree;
namespace loc = boost::locale;

//----------------------------------------------------------------------
//! \brief Constructor
//----------------------------------------------------------------------
CFileConverter::CFileConverter()
        : mOpt(0)
        , mSrcFilePath(L"")
        , mDstFilePath(L"")
        , mStopId(0)
        , mRouteName(L"")
        , mCoordsFormat(0)
{
}

//----------------------------------------------------------------------
//! \brief Parse the command line arguments
//! \param argc Argument count
//! \param argv Argument list
//! \return 0 if successful
//----------------------------------------------------------------------
int CFileConverter::ParseCommandLine( int argc, wchar_t *argv[] )
{
    try
    {
        po::options_description desc( "Allowed Options" );
        desc.add_options()
            ("help,H", "produce help message")
            ("conversion_type", po::value<uint32>(&mOpt), "Type of file conversion\nValues:\n"
                                                "   1 -- PSR to CSV\n"
                                                "   2 -- CSV to PSR\n"
                                                "   3 -- GPX to PSR")
            ("src_file", po::wvalue<wstring>(), "Full path name of the source file")
            ("dst_file", po::wvalue<wstring>(), "Full path name of the destination file")
            ("stop_ID,I", po::value<uint32>(), "Stop ID (Required if CSV to PSR or GPX to PSR)")
            ("route_name,R", po::wvalue<wstring>(), "Desired name for the route (Required if CSV "
                                                "to PSR)")
            ("src_coords,S", po::value<short>(), "Format of the coordinates of the source file "
                                               "(Required if CSV to PSR)"
                                               "\nValues:\n   1 -- degrees\n   2 -- semicircles")
            ("dst_coords,D", po::value<short>(), "Desired format for the coordinates of the converted"
                                               " file (Required if PSR to CSV)"
                                               "\nValues:\n   1 -- degrees\n   2 -- semicircles")
        ;
        po::positional_options_description pd;
        pd.add("conversion_type", 1).add("src_file", 1).add("dst_file", 1);

        po::variables_map vm;
        po::store( po::wcommand_line_parser( argc, argv ).options(desc).positional(pd).run(), vm );
        po::notify(vm);

        if( vm.count( "help" ) )
        {
            cout << desc << "\n";
            return ERROR;
        }
        if( mOpt > 0 && mOpt < 4 )
        {
            string type;
            if( mOpt == 1 ) type = "PSR to CSV";
            else if( mOpt == 2 ) type = "CSV to PSR";
            else type = "GPX to PSR";
            cout << "Conversion_type was set to " << type << ".\n";
        }
        else
        {
            cout << "Conversion_type must be in the range [1 to 3].\n";
            return ERROR;
        }
        if( vm.count( "src_file" ) )
        {
            mSrcFilePath = vm["src_file"].as< wstring >();
            wcout << L"Source File: "
                  << mSrcFilePath << endl;
        }
        else
        {
            cout << "Error: src_file was not set.\n";
            return ERROR;
        }
        if( vm.count( "dst_file" ) )
        {
            mDstFilePath = vm["dst_file"].as< wstring >();
            wcout << L"Destination File: "
                  << mDstFilePath << endl;
        }
        else
        {
            cout << "Error: dst_file was not set.\n";
            return ERROR;
        }
        if( mOpt == PSRtoCSV )
        {
            if( vm.count( "dst_coords" ) )
            {
                mCoordsFormat = vm["dst_coords"].as< short >();
                return SUCCESS;
            }
            else
            {
                cout << "Error: PSR to CSV was chosen but dst_coords was not set.\n";
                return ERROR;
            }
        }
        else if( mOpt == GPXtoPSR )
        {
            if( vm.count( "stop_ID" ) )
            {
                mStopId = vm["stop_ID"].as< uint32 >();
                cout << "Stop ID: " << mStopId << "\n";
                return SUCCESS;
            }
            else
            {
                cout << "Error: GPX to PSR was chosen but stop_ID was not set.\n";
                return ERROR;
            }
        }
        else if( mOpt == CSVtoPSR )
        {
            if( vm.count( "stop_ID" ) && vm.count( "route_name" ) && vm.count( "src_coords" ) )
            {
                mStopId = vm["stop_ID"].as< uint32 >();
                mRouteName = vm["route_name"].as< wstring >();
                mCoordsFormat = vm["src_coords"].as< short >();
                cout << "Stop ID: " << mStopId << "\n";
                wcout << L"Route Name: " << mRouteName << L"\n";
                return SUCCESS;
            }
            else
            {
                cout << "Error: CSV to PSR was chosen but not all required arguments were set.\n";
                return ERROR;
            }
        }
        else
        {
            return ERROR; // Should never get here
        }
    }
    catch( exception& e )
    {
        cerr << "Error: " << e.what() << "\n";
        return ERROR;
    }
}

//----------------------------------------------------------------------
//! \brief Attempt to convert a CSV file of coordinates to a PSR File
//! \details If any error is discovered, a message box will prompt the
//!     operator of the issue. The output file is specified by the user.
//----------------------------------------------------------------------
int CFileConverter::ConvertCSVtoPSR()
{
    wifstream csvFile( mSrcFilePath.c_str(), ios_base::in );
    if( !csvFile.good() )
    {
        wcerr << L"File Conversion Error\n\nFailed to open the source file: "
             << mSrcFilePath << L"\n";
        return ERROR;
    }

    ofstream psrFile( mDstFilePath.c_str(), ios_base::out | ios_base::binary );
    if( !psrFile.good() )
    {
        wcerr << L"File Conversion Error\n\nFailed to create the destination file: "
             << mDstFilePath << L"\n";
        csvFile.close();
        return ERROR;
    }

    WritePSRHeader( psrFile );

    sc_position_type pos; // protocol only supports sending sint32 (semicircles).
                          // floating point numbers (degrees) will be converted.
    wstring pointName;
    id_type pointType;
    int result = SUCCESS;
    bool firstPt = true;
    bool isShaping = true;

    while ( csvFile.good() && !csvFile.eof() )
    {
        // read from csv data
        vector<wstring> row;  // row[ 0:name, 1:lat, 2:lon, 3:type ]
        wstring line, cell;

        getline( csvFile, line );
        if( line.length() == 0 )
        {
            continue; // empty line (typically trailing lines at end of file)
        }
        wstringstream lineStream( line );

        // parse csv line
        while( getline( lineStream, cell, L',' ) )
        {
            row.push_back( cell );
        }
        if( row.size() != 4 )
        {
           cerr << "File Conversion Error\n\nError found in the CSV File."
                   " Each row in the file must contain 4 cells:\n\n\tPoint Name"
                   " (if shaping leave blank)\n\tLatitude\n\tLongitude\n"
                   "\tPoint Type (1=Destination, 2=Shaping).\n";
            result = ERROR;
            break;
        }

        // Extract row data
        double lat, lon = 0;
        pointName = row[0];
        wistringstream iss( row[1] );
        iss >> lat;
        iss.clear(); iss.str( row[2] );
        iss >> lon;
        if( mCoordsFormat == DEGREES )
        {
            pos.lat = convert_degrees_to_semicircles( lat );
            pos.lon = convert_degrees_to_semicircles( lon );
        }
        else
        {
            pos.lat = (sint32) lat;
            pos.lon = (sint32) lon;
        }
        pointType = (uint8)wcstol( row[3].c_str(), NULL, 0 );

        // Destination pt
        if( pointType == 1 )
        {
            if( !pointName.empty() )
            {
                psrFile.write( ( const char *) &pos.lat, sizeof( pos.lat ) );
                psrFile.write( ( const char *) &pos.lon, sizeof( pos.lon ) );
                psrFile.write( "\0", 1 );
                string utf8_pointName = loc::conv::utf_to_utf<char>( pointName );
                psrFile.write( (const char*) utf8_pointName.c_str(), utf8_pointName.size() );
                psrFile.write( "\0", 1 );
                isShaping = false;
                firstPt = false;
            }
            else
            {
                cerr << "File Conversion Error\n\nError found in the CSV File."
                        " Destination Points must have a valid name.\n";
                result = ERROR;
                break;
            }
        }
        // Shaping pt
        else if( pointType == 2 )
        {
            // First point cannot be a shaping point
            if( firstPt )
            {
                cerr << "File Conversion Error\n\nError found in the CSV File."
                        " First point in the file is a Shaping point. The first and"
                        " last points in the file must be Destination Points.\n";
                result = ERROR;
                break;
            }
            psrFile.write( ( const char *) &pos.lat, sizeof( pos.lat ) );
            psrFile.write( ( const char *) &pos.lon, sizeof( pos.lon ) );
            psrFile.write( "\1", 1 );
            isShaping = true;
        }
        else
        {
            cerr << "File Conversion Error\n\nError found in the CSV File."
                    " Invalid Point Type. Types must be either 1 (Destination)"
                    " or 2 (Shaping).\n";
            result = ERROR;
            break;
        }
    }
    if( result == SUCCESS && isShaping )
    {
        // Last point in file was a shaping pt, which is not allowed
        cerr << "File Conversion Error\n\nError found in the CSV File."
                " Final point in the file was a Shaping point. The first and"
                " last points in the file must be Destination Points.\n";
        result = ERROR;
    }

    csvFile.close();
    psrFile.close();
    // File was not converted successfully so delete it.
    if( result == ERROR )
    {
        DeleteDstFile();
    }

    return result;
}

//----------------------------------------------------------------------
//! \brief Attempt to convert a PSR file to a CSV File of coordinates
//! \details If any error is discovered, a message box will prompt the
//!     operator of the issue. The output file is specified by the user.
//----------------------------------------------------------------------
int CFileConverter::ConvertPSRtoCSV()
{
    int result = SUCCESS;
    const uint32 HEADER_SIZE = 14;  // Minus Stop Text
    const uint32 STOP_SIGNATURE = 0x2B525350;

    ifstream psrFile( mSrcFilePath.c_str(), ios_base::binary | ios_base::in );
    if( !psrFile.good() )
    {
        wcerr << L"File Conversion Error\n\nFailed to open the source file: "
             << mSrcFilePath << "\n";
        return ERROR;
    }

    wofstream csvFile( mDstFilePath.c_str(), ios_base::out | ios_base::trunc );
    std::locale utf8Locale(std::locale(), new po::detail::utf8_codecvt_facet());
    csvFile.imbue( utf8Locale ); // enforce that the file is encoded in UTF-8
    if( !csvFile.good() )
    {
        wcerr << L"File Conversion Error\n\nFailed to create the destination file: "
             << mDstFilePath << "\n";
        psrFile.close();
        return ERROR;
    }
    // Copy data into a vector
    vector<unsigned char> psrBuffer( ( istreambuf_iterator<char>( psrFile ) ),
                                     ( istreambuf_iterator<char>( ) ) );
    vector<unsigned char>::iterator it;
    it = psrBuffer.begin();

    // Verify the file has at least enough data to contain the header
    //  and that it conatins the PSR+ notation
    if( psrBuffer.size() <= HEADER_SIZE ||
            STOP_SIGNATURE != *reinterpret_cast<uint32*>(&*it) )
    {
        cerr << "File Conversion Error\n\nThe PSR File header is incorrectly formatted.";
        psrFile.close();
        csvFile.close();
        return ERROR;
    }

    // Advance iterator to the beginning of Stop Text
    //   PSR+ header     (4 bytes)
    //   File format     (2 bytes)
    //   Timestamp       (4 bytes)
    //   Unique ID       (4 bytes)
    //   Stop Text       (Max 200 bytes)
    advance(it, HEADER_SIZE);

    // Iterate over the Stop text until we find the null character.
    while( it != psrBuffer.end() && *it++ != NULL )
    {
        if( it == psrBuffer.end() )
        {
            cerr << "File Conversion Error\n\nThe PSR File header is incorrectly formatted.";
            result = ERROR;
            break;
        }
    }
    double_position_type pos;
    id_type pointType = 0;

    // Main data loop
    while( it != psrBuffer.end() )
    {
        wstring csvRow(L"");
        if( mCoordsFormat == DEGREES )
        {
            pos.lat = convert_semicircles_to_degrees( *reinterpret_cast<sint32*>(&*it) );
            advance( it, sizeof(sint32) );
            pos.lon = convert_semicircles_to_degrees( *reinterpret_cast<sint32*>(&*it) );
            advance( it, sizeof(sint32) );
        }
        else
        {
            pos.lat = *reinterpret_cast<sint32*>(&*it);
            advance( it, sizeof(sint32) );
            pos.lon = *reinterpret_cast<sint32*>(&*it);
            advance( it, sizeof(sint32) );
        }
        // First byte: point type
        pointType = *it++;
        // If destination point, get the point name
        if( pointType == 0 )
        {
            pointType = 1;  // PSR stores dst points as 0, CSV format is 1
            stringstream ss;
            while( *it != NULL )
            {
                ss << *it++;
            }
            csvRow += loc::conv::to_utf<wchar_t>( ss.str(), "UTF-8" );
            // Advance iterator to start of next data item
            if( it != psrBuffer.end() )
            {
                it++;
            }
        }
        else if( pointType == 1 )
        {
            pointType = 2; // PSR stores shaping points as 1, CSV format is 2
        }
        else
        {
            cerr << "File Conversion Error\n\nThe PSR File contained an unkown Point Type.";
            result = ERROR;
            break;
        }
        // Concatenate the CSV row
        wstringstream ss;
        ss.precision(8);
        csvRow += L",";
        if( mCoordsFormat == DEGREES )
        {
            ss << pos.lat;
            csvRow += ss.str(); ss.str(L"");
            csvRow += L",";
            ss << pos.lon;
            csvRow += ss.str(); ss.str(L"");
            csvRow += L",";
        }
        else
        {
            ss << (sint32)pos.lat;
            csvRow += ss.str(); ss.str(L"");
            csvRow += L",";
            ss << (sint32)pos.lon;
            csvRow += ss.str(); ss.str(L"");
            csvRow += L",";
        }
        ss << static_cast<uint32>(pointType);
        csvRow += ss.str();

        // Write the CSV row to the file
        csvFile << csvRow << endl;
    }
    psrFile.close();
    csvFile.close();
    // File was not converted successfully so delete it.
    if( result == ERROR )
    {
        DeleteDstFile();
    }

    return result;
}

//----------------------------------------------------------------------
//! \brief Attempt to convert a GPX file to a PSR File
//! \details If any error is discovered, a message box will prompt the
//!     operator of the issue. The output file is specified by the user.
//----------------------------------------------------------------------
int CFileConverter::ConvertGPXtoPSR()
{
    pt::wptree tree;
    std::wifstream f( mSrcFilePath.c_str() );
    std::locale utf8Locale(std::locale(), new po::detail::utf8_codecvt_facet());
    f.imbue( utf8Locale );  // enforce that the file be encoded in UTF-8
    pt::read_xml( f, tree );
    boost::optional< pt::wptree& > rte = tree.get_child_optional( L"gpx.rte" );
    if( !rte )
    {
        cerr << "File Conversion Error\n\nNo route defined in the GPX file.\n";
        return ERROR;
    }
    try
    {
        mRouteName = tree.get<std::wstring>( L"gpx.rte.name" );
    }
    catch( pt::ptree_bad_path )
    {
        cout << "File Conversion Warning! The specified GPX file does not contain "
             "a route name.\n";
    }

    ofstream psrFile( mDstFilePath.c_str(), ios_base::out | ios_base::binary );
    if( !psrFile.good() )
    {
        wcerr << L"File Conversion Error\n\nFailed to create the destination file: "
             << mDstFilePath << endl;
        return ERROR;
    }

    WritePSRHeader( psrFile );
    sc_position_type pos; // protocol only supports sending sint32 (semicircles).
    wstring pointName(L"");
    int result = SUCCESS;
    try
    {
        BOOST_FOREACH( pt::wptree::value_type &rtept, tree.get_child( L"gpx.rte" ) )
        {
            if( rtept.first == L"rtept" )
            {
                pointName = rtept.second.get<std::wstring>( L"name", L"" );
                // GPX lats and longs are always stored in decimal degrees
                pos.lat = convert_degrees_to_semicircles( rtept.second.get<double>( L"<xmlattr>.lat", 0 ) );
                pos.lon = convert_degrees_to_semicircles( rtept.second.get<double>( L"<xmlattr>.lon", 0 ) );
                // Write the stop data to the PSR file
                psrFile.write( ( const char *) &pos.lat, sizeof( pos.lat ) );
                psrFile.write( ( const char *) &pos.lon, sizeof( pos.lon ) );
                psrFile.write( "\0", 1 ); // GPX only contains destination points
                string utf8_pointName = loc::conv::utf_to_utf<char>( pointName );
                psrFile.write( ( const char *)utf8_pointName.c_str(), utf8_pointName.size() );
                psrFile.write( "\0", 1 );
            }
        }
    }
    catch( pt::ptree_bad_path )
    {
        cerr << "File Conversion Error\n\nA GPX route point was incorrectly formatted\n";
        result = ERROR;
    }
    psrFile.close();
    // File was not converted successfully so delete it.
    if( result == ERROR )
    {
        DeleteDstFile();
    }

    return result;
}

//----------------------------------------------------------------------
//! \brief Write the standard PSR header to the output stream
//! \details The timestamp written to the file is the current time
//----------------------------------------------------------------------
void CFileConverter::WritePSRHeader( ofstream &psrFile )
{
    if( psrFile.is_open() )
    {
        short ver = 0;
        time_type origination_time = get_current_garmin_time();

        // Write PSR Header info
        psrFile.write( "PSR+", sizeof( uint32 ) );
        psrFile.write( ( const char* ) &ver, sizeof( ver ) );
        psrFile.write( ( const char* ) &origination_time, sizeof( origination_time ) );
        psrFile.write( ( const char* ) &mStopId, sizeof( mStopId ) );
        string utf8_routeName = loc::conv::utf_to_utf<char>( mRouteName );
        psrFile.write( (const char*) utf8_routeName.c_str(), utf8_routeName.size() );
        psrFile.write( "\0", 1 );
        psrFile.flush();
    }
}

//----------------------------------------------------------------------
//! \brief Delete the destination file
//! \details Issued when the conversion fails. We don't want to leave a
//!     corrupt file leftover.
//----------------------------------------------------------------------
void CFileConverter::DeleteDstFile()
{
    try
    {
        _wremove( mDstFilePath.c_str() );
    }
    catch ( exception ex )
    {
        wcout << L"\nFile " << mDstFilePath << L" could not be removed\n";
    }
}