/*********************************************************************
*
*   MODULE NAME:
*       Utility.h
*
* Copyright 2008-2016 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/

//! Note: This file contains some useful content borrowed from the FMC
//!       application. In order to decouple the Fleet File Utility from
//!       the Windows platform, there can be no file dependency between
//!       programs. This enables the Utility to compile and run as a
//!       stand-alone app on either the Windows or Linux platform.

#define ERROR  -1
#define SUCCESS 1

//! \brief Number of seconds in one week
#define SECONDS_PER_WEEK        ( 60 * 60 * 24 * 7 )
//! \brief Number of seconds in one day
#define SECONDS_PER_DAY         ( 60 * 60 * 24 )
//! \brief Number of seconds in one hour
#define SECONDS_PER_HOUR        ( 60 * 60 )
//! \brief Number of seconds in one minute
#define SECONDS_PER_MINUTE      ( 60 )
//! \brief Number of days in a non-leap year
#define DAYS_IN_1_YEAR          ( 365 )

#define DEGREES 1
#define SEMICIRCLES 2

//! 8-bit signed integer
typedef signed char             sint8;
//! 16-bit signed integer
typedef signed short int        sint16;
//! 32-bit signed integer
typedef signed long int         sint32;

//! 8-bit unsigned integer
typedef unsigned char           uint8;
//! 16-bit unsigned integer
typedef unsigned short int      uint16;
//! 32-bit unsigned integer
typedef unsigned long int       uint32;

//! \brief 32-bit IEEE-format floating point data.
//! (1 sign bit, 8 exponent bits, and 23 mantissa bits)
typedef float                   float32;

//! \brief 64-bit IEEE-format floating point data.
//! (1 sign bit, 11 exponent bits, and 52 mantissa bits)
typedef double                  float64;

//! \brief Garmin packet ID.
//! \see id_enum for supported values
typedef uint8 id_type;

//! \brief Absolute time (number of seconds since 12/31/1989 12:00 am UTC)
//!
//! \details The time_type is used to indicate an absolute time. It is an
//! unsigned 32-bit integer and its value is the number of seconds since
//! 12:00 am, December 31, 1989 UTC. A hex value of 0xFFFFFFFF represents
//! an invalid time, and the client will ignore the time.
typedef uint32 time_type;

//! \brief Indicates a latitude and longitude in semicircles.
//! \details Indicates a 2D position in semicircles, where 2^31
//!     semicircles equal 180 degrees. North latitudes and East
//!     longitudes are indicated with positive numbers; South latitudes
//!     and West longitudes are indicated with negative numbers.  All
//!     positions are given in WGS-84.
struct sc_position_type
    {
    sint32      lat;         //!< latitude in semicircles
    sint32      lon;         //!< longitude in semicircles
    };

//! \brief Indicates a latitude and longitude in radians.
//! \details Indicates a 2D position in radians. North latitudes and
//!     East longitudes are indicated with positive numbers; South
//!     latitudes and West longitudes are indicated with negative
//!     numbers.  All positions are given in WGS-84.
struct double_position_type
    {
    float64     lat;         //!< latitude in radians, positive is north
    float64     lon;         //!< longitude in radians, positive is east
    };

//! Helper static utility functions
//----------------------------------------------------------------------
//! \brief Converts a latitude/longitude from semicircles to degrees.
//! \param aSemicircles The latitude/longitude in semicircles
//! \return The latitude/longitude in degrees
//----------------------------------------------------------------------
static inline double convert_semicircles_to_degrees( sint32 aSemicircles )
{
    return (double)aSemicircles * 90 / ( 1 << 30 );
}
//----------------------------------------------------------------------
//! \brief Converts a latitude/longitude from degrees to semicircles.
//! \details 2^31 semicircles equal 180 degrees.  The conversion uses
//!     the equivalent 2^30 semicircles per 90 degrees to avoid
//!     overflow.
//! \param aDegrees The latitude/longitude in degrees
//! \return The latitude/longitude in semicircles.
//----------------------------------------------------------------------
static inline sint32 convert_degrees_to_semicircles( double aDegrees )
{
    return (sint32)( aDegrees * (double)( ( 1 << 30 ) / 90 ) );
}

time_type get_current_garmin_time();

time_type get_garmin_time( tm aTimeTm );