/*********************************************************************
*
*   MODULE NAME:
*       SensorListItem.cpp
*
*   Copyright 2014 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/

#include "SensorListItem.h"
#include <sstream>

#if( FMI_SUPPORT_A617 )

//----------------------------------------------------------------------
//! \brief Constructor
//----------------------------------------------------------------------
SensorListItem::SensorListItem() : ClientListItem()
{
    numUpdates = 0;
}

void SensorListItem::readFromStream( std::istream &aStream )
{
    fmi_sensor_config_request config;
    fmi_sensor_update_request update;

    if( !aStream.eof() )
    {
        aStream >> config.change_id;
        aStream.get();
    }
    if( !aStream.eof() )
    {
        uint8 buffer;
        aStream >> buffer;
        config.display_index = buffer;
        aStream.get();
    }
    if( !aStream.eof() )
    {
        aStream >> config.icon;
        aStream.get();
    }
    if( !aStream.eof() )
    {
        aStream >> numUpdates;
        aStream.get();
    }
    if( numUpdates > 0 )
    {
        if( !aStream.eof() )
        {
            aStream >> update.change_id;
            aStream.get();
        }
        if( !aStream.eof() )
        {
            aStream >> update.unique_id;
            aStream.get();
        }
        if( !aStream.eof() )
        {
            uint8 buffer;
            aStream >> buffer;
            update.severity = buffer;
            aStream.get();
        }
        if( !aStream.eof() )
        {
            int buffer;
            aStream >> buffer;
            update.play_sound = buffer;
            aStream.get();
        }
        if( !aStream.eof() )
        {
            int buffer;
            aStream >> buffer;
            update.record_sensor = buffer;
            aStream.get();
        }
        if( !aStream.eof() )
        {
            aStream.getline(update.status, sizeof(update.status), '-');
        }
        if( !aStream.eof() )
        {
            aStream.getline(update.description, sizeof(update.status), '-');
        }
    }

    if( !aStream.eof() )
    {
        aStream >> config.unique_id;
        aStream.get();
    }
    if( !aStream.eof() )
    {
        aStream >> config.name;
    }

    setConfig( config );

    //config's unique id has to be set before this happens.
    if( numUpdates > 0 )
    {
        setUpdate( update );
    }

    setCurrentName( CString( config.name ) );
    setId( config.unique_id );

    //make sure it's not a sensor with junk data
    if( config.icon < 52428 )
        {
        setValid();
        }
}

void SensorListItem::writeToStream( std::ofstream &aStream )
{
        std::ostringstream os;
        os << (uint32)getChangeId();
        aStream << os.str().c_str();
        aStream << "-";
        os.clear();
        os.str("");

        os << (uint8)getDisplayIndex();
        aStream << os.str().c_str();
        aStream << "-";
        os.clear();
        os.str("");

        os << getIcon();
        aStream << os.str().c_str();
        aStream << "-";
        os.clear();
        os.str("");

        os << numUpdates;
        aStream << os.str().c_str();
        aStream << "-";
        os.clear();
        os.str("");

        if( numUpdates > 0 )
        {
            os << (uint32)mUpdate.change_id;
            aStream << os.str().c_str();
            aStream << "-";
            os.clear();
            os.str("");

            os << (uint32)mUpdate.unique_id;
            aStream << os.str().c_str();
            aStream << "-";
            os.clear();
            os.str("");

            os << (uint8)mUpdate.severity;
            aStream << os.str().c_str();
            aStream << "-";
            os.clear();
            os.str("");

            os << (int)getPlaySound();
            aStream << os.str().c_str();
            aStream << "-";
            os.clear();
            os.str("");

            os << (int)getRecordSensor();
            aStream << os.str().c_str();
            aStream << "-";
            os.clear();
            os.str("");

            aStream << getStatus();
            aStream << "-";

            aStream << getDescription();
            aStream << "-";
        }

        ClientListItem::writeToStream( aStream );
}

//----------------------------------------------------------------------
//! \brief Get the number of updates submitted
//! \return the numUpdates data
//----------------------------------------------------------------------
uint32 SensorListItem::getNumUpdates() const
{
    return numUpdates;
}

//----------------------------------------------------------------------
//! \brief Get the change id
//! \return the id of the sensor
//----------------------------------------------------------------------
uint32 SensorListItem::getChangeId() const
{
    return mConfig.change_id;
}

//----------------------------------------------------------------------
//! \brief Get the unique id
//! \return the unique id of the config alert popup
//----------------------------------------------------------------------
uint32 SensorListItem::getUniqueId() const
{
    return mConfig.unique_id;
}

//----------------------------------------------------------------------
//! \brief Get icon
//! \return the icon of the sensor
//----------------------------------------------------------------------
uint16 SensorListItem::getIcon() const
{
    return mConfig.icon;
}

//----------------------------------------------------------------------
//! \brief Get display index
//! \return the display index of the sensor
//----------------------------------------------------------------------
uint8 SensorListItem::getDisplayIndex() const
{
    return mConfig.display_index;
}

//----------------------------------------------------------------------
//! \brief Get name
//! \return the name of the sensor
//----------------------------------------------------------------------
const char* SensorListItem::getName() const
{
    return &mConfig.name[0];
}

//----------------------------------------------------------------------
//! \brief Get the severity
//! \return the severity of the last update
//----------------------------------------------------------------------
uint8 SensorListItem::getSeverity() const
{
    return mUpdate.severity;
}

//----------------------------------------------------------------------
//! \brief Get the play sound setting
//! \return the play sound setting of the last update
//----------------------------------------------------------------------
boolean SensorListItem::getPlaySound() const
{
    return mUpdate.play_sound;
}

//----------------------------------------------------------------------
//! \brief Get the record sensor setting
//! \return the record sensor setting of the last update
//----------------------------------------------------------------------
boolean SensorListItem::getRecordSensor() const
{
    return mUpdate.record_sensor;
}

//----------------------------------------------------------------------
//! \brief Get the status
//! \return the status of the last update
//----------------------------------------------------------------------
const char* SensorListItem::getStatus() const
{
    return &mUpdate.status[0];
}

//----------------------------------------------------------------------
//! \brief Get the description
//! \return the description of the last update
//----------------------------------------------------------------------
const char* SensorListItem::getDescription() const
{
    return &mUpdate.description[0];
}

//----------------------------------------------------------------------
//! \brief Set the sensor
//! \param aConfig The config that serves as our sensor
//----------------------------------------------------------------------
void SensorListItem::setConfig
    (
    fmi_sensor_config_request aConfig
    )
{
    mConfig = aConfig;
}

//----------------------------------------------------------------------
//! \brief Set the update
//! \param aUpdate The update to this sensor
//----------------------------------------------------------------------
void SensorListItem::setUpdate
    (
    fmi_sensor_update_request aUpdate
    )
{
    if( getUniqueId() == aUpdate.unique_id )
    {
        mUpdate = aUpdate;
        numUpdates++;
    }
}

//----------------------------------------------------------------------
//! \brief Set the parent
//! \param aParent The FileBackedMap that contains this item
//----------------------------------------------------------------------
void SensorListItem::setParent
    (
    FileBackedMap<SensorListItem> * aParent
    )
{
    mParent = aParent;
}

//----------------------------------------------------------------------
//! \brief Set the current name of the sensor
//! \param aName The new name of the sensor
//----------------------------------------------------------------------
void SensorListItem::setCurrentName
    (
    CString aName
    )
{
    mCurrentName = aName;
}
//----------------------------------------------------------------------
//! \brief Set dispolay index in the list
//! \param index, the new index
//----------------------------------------------------------------------
void SensorListItem::setDisplayIndex
    (
    uint8 index
    )
{
    mConfig.display_index = index;
}


#endif