/*********************************************************************
*
*   MODULE NAME:
*       TcpIpPort.cpp
*
*   Copyright 2016 by Garmin Ltd. or its subsidiaries.
*---------------------------------------------------------------------
* $NoKeywords$
*********************************************************************/

#include "stdafx.h"
#include "TcpIpPort.h"
#include "LinkLayer.h"
#include "Logger.h"
#include "GarminLinkLayer.h"

#include <afx.h>
#include <afxext.h>
#include <iostream>

#define RX_BUFFER_SIZE       ( 1024 )

bool TcpIpPort::initTcpPort()
{
    if( AfxSocketInit() == FALSE)
    {
        sInstance->recordErrorText( _T( "Error calling AfxSocketInit()" ) );
        return false;
    }

    if( sInstance )
    {
        sInstance->close();
    }

    // Create a TCP/IP virtual serial port and connect it to the link layer
    sInstance = new TcpIpPort();
    GarminLinkLayer::getInstance()->resetPhysicalLayer( sInstance );

    // Only clear the log if it isn't already open.
    // Reinitializing the com port shouldn't clear the log.
    if( !Logger::isLogOpen() )
        Logger::clearLog();

    return ( (TcpIpPort *) sInstance )->init();
}

bool TcpIpPort::init()
{
    socketIsOpen = false;

    clientSocket.Create();
    CString localHost("127.0.0.1");
    if(clientSocket.Connect(localHost, 6000))
    {
        socketIsOpen = true;
    }
    else
    {
        recordErrorText( _T( "Error calling clientSocket.Connect(localHost, 6000)" ) );
    }
    return socketIsOpen;
}

//----------------------------------------------------------------------
//! \brief Constructor
//----------------------------------------------------------------------
TcpIpPort::TcpIpPort()
{
}

//----------------------------------------------------------------------
//! \brief Destructor.  Close the serial port.
//----------------------------------------------------------------------
TcpIpPort::~TcpIpPort()
{
    if( mLinkLayer )
    {
        mLinkLayer->setPhysicalLayer( NULL );
    }
    close();
}

//----------------------------------------------------------------------
//! \brief Receive and process any data
//----------------------------------------------------------------------
void TcpIpPort::pumpRx()
{
    uint8                   readBuffer[ RX_BUFFER_SIZE ];
    int                     readSize;

    if( socketIsOpen )
    {
        // Get the number of bytes that can be read without blocking
        DWORD availableBytes = 0;
        if( clientSocket.IOCtl( FIONREAD, &availableBytes) == 0 )
        {
            recordErrorText( _T( "Error polling socket for available data." ) );
            close();
            return;
        }

        if( availableBytes > 0 )
        {
            if( availableBytes > RX_BUFFER_SIZE )
            {
                availableBytes = RX_BUFFER_SIZE;
            }
            readSize = clientSocket.Receive(readBuffer, availableBytes);
            if (readSize > 0)
            {
                // Push the received bytes up to the link layer
                if( mLinkLayer )
                {
                    mLinkLayer->rx( readBuffer, readSize );
                }
            }
        }
    }
}

//----------------------------------------------------------------------
//! \brief Close the TCP virtual COM port if one is in use.
//----------------------------------------------------------------------
void TcpIpPort::close()
{
    if( isOpen() )
    {
        clientSocket.Close();
    }
    socketIsOpen = false;
}

//----------------------------------------------------------------------
//! \brief Transmit bytes on the serial port.
//! \param aData The bytes to transmit
//! \param aSize The number of bytes to transmit
//! \return true if the bytes were transmitted, false otherwise
//----------------------------------------------------------------------
bool TcpIpPort::tx
    (
    uint8  * aData,
    uint16   aSize
    )
{
    if( socketIsOpen )
    {
        if( SOCKET_ERROR != clientSocket.Send(aData, aSize) )
        {
            return true;
        }
        else
        {
            recordErrorText( _T( "Error sending data on socket" ) );
            close();
        }
    }

    return false;
}

//----------------------------------------------------------------------
//! \brief Get the name of the serial port that is open.
//! \return The name of the serial port that is open, or an empty
//!     string if the port is closed.
//----------------------------------------------------------------------
const CString& TcpIpPort::getPortName() const
{
    return mPortName;
}

//----------------------------------------------------------------------
//! \brief Indicate whether the port is open.
//! \return true if the serial port is open, false otherwise.
//----------------------------------------------------------------------
bool TcpIpPort::isOpen() const
{
    return socketIsOpen;
}

//----------------------------------------------------------------------
//! \brief Set the baud rate.
//! \param aBaudRate The new baud rate for the port, in bps.
//! \
//----------------------------------------------------------------------
bool TcpIpPort::setBaudRate
    (
    uint32 aBaudRate
    )
{
    mBaudRate = aBaudRate;
    return true;
}
//----------------------------------------------------------------------
//! \brief Return the last successfully applied baud rate
//! \
//----------------------------------------------------------------------
uint32 TcpIpPort::getBaudRate() const
    {
    return mBaudRate;
    }
